@@ -1,3 +1,5 @@
+use std::borrow::Borrow;
+
 use crate::internal::*;
 use crate::model::{TypedModel, TypedNode};
 use crate::ops::identity::Identity;
@@ -28,6 +30,7 @@ pub enum AxisOp {
     Move(usize, usize),
     Reshape(usize, TVec<TDim>, TVec<TDim>),
 }
+
 use AxisOp::*;
 
 impl PartialEq for AxisOp {
@@ -58,6 +61,45 @@ impl AxisOp {
         }
     }
 
+    pub fn simplify(&self) -> TVec<AxisOp> {
+        match self.canonical().borrow() {
+            Reshape(_, from, to) if from == to => tvec!(),
+            Reshape(at, from, to) if to.len() == 0 => tvec!(Rm(*at); from.len()),
+            Reshape(at, from, to) if from.len() == 0 => tvec!(Add(*at); to.len()),
+            Reshape(at, from, to) if from[0] == to[0] => {
+                Reshape(at + 1, from[1..].into(), to[1..].into()).simplify()
+            }
+            Reshape(at, from, to) if from[from.len() - 1] == to[to.len() - 1] => {
+                Reshape(*at, from[..from.len() - 1].into(), to[..to.len() - 1].into()).simplify()
+            }
+            Reshape(at, from, to) if from[0] == 1.to_dim() => std::iter::once(Rm(*at))
+                .chain(Reshape(*at, from[1..].into(), to.clone()).simplify().into_iter())
+                .collect(),
+            Reshape(at, from, to) if to[0] == 1.to_dim() => {
+                Reshape(*at, from.clone(), to[1..].into())
+                    .simplify()
+                    .into_iter()
+                    .chain(std::iter::once(Add(*at)))
+                    .collect()
+            }
+            Reshape(at, from, to) if from[from.len() - 1] == 1.to_dim() => std::iter::once(Rm(at
+                + from.len()
+                - 1))
+            .chain(Reshape(*at, from[..from.len() - 1].into(), to.clone()).simplify().into_iter())
+            .collect(),
+            Reshape(at, from, to) if to[to.len() - 1] == 1.to_dim() => {
+                std::iter::once(Add(at + from.len()))
+                    .chain(
+                        Reshape(*at, from.clone(), to[..to.len() - 1].into())
+                            .simplify()
+                            .into_iter(),
+                    )
+                    .collect()
+            }
+            other => tvec!(other.clone()),
+        }
+    }
+
     pub fn transform_axis(&self, axis: usize) -> Option<usize> {
         match self.canonical().as_ref() {
             Add(ix) => Some(axis + (axis >= *ix) as usize),
@@ -222,7 +264,7 @@ impl AxisOp {
     pub fn change_shape_array<D: DimLike>(
         &self,
         shape: &mut TVec<D>,
-        broadcast: bool,
+        broadcasting: bool,
     ) -> TractResult<()> {
         use std::convert::TryInto;
         match self.canonical().as_ref() {
@@ -235,24 +277,34 @@ impl AxisOp {
                 shape.insert(*to, axis);
             }
             Reshape(at, from, to) => {
-                if shape.len() < from.len() + *at
-                    || (!broadcast
-                        && shape.iter().skip(*at).zip(from.iter()).any(|(s, f)| &s.to_dim() != f))
+                if shape.len() >= from.len() + *at
+                    && tract_itertools::izip!(shape.iter().skip(*at), from)
+                        .all(|(shape, spec)| shape.to_dim() == *spec)
                 {
+                    for _ in from {
+                        shape.remove(*at);
+                    }
+                    for d in to.iter().rev() {
+                        shape.insert(*at, d.try_into()?);
+                    }
+                } else if broadcasting
+                    && shape.iter().skip(*at).take(from.len()).all(|d| d.to_dim() == 1.to_dim())
+                {
+                    for _ in from {
+                        shape.remove(*at);
+                    }
+                    for _ in to.iter().rev() {
+                        shape.insert(*at, 1.into());
+                    }
+                } else {
                     bail!("Incompatible reshape for shape {:?} and {:?}", shape, self);
                 }
-                for _ in from {
-                    shape.remove(*at);
-                }
-                for d in to.iter().rev() {
-                    shape.insert(*at, d.try_into()?);
-                }
             }
         }
         Ok(())
     }
 
-    pub fn change_shape(&self, shape: &mut ShapeFact, broadcast: bool) -> TractResult<()> {
+    pub fn change_shape(&self, shape: &mut ShapeFact, broadcasting: bool) -> TractResult<()> {
         match self.canonical().as_ref() {
             Add(ix) => shape.insert_axis(*ix),
             Rm(ix) => {
@@ -266,7 +318,7 @@ impl AxisOp {
             }
             _ => {
                 let mut array = shape.to_tvec();
-                self.change_shape_array(&mut array, broadcast)?;
+                self.change_shape_array(&mut array, broadcasting)?;
                 let mut new_shape = ShapeFact::from_dims(array);
                 std::mem::swap(shape, &mut new_shape);
                 Ok(())
@@ -279,16 +331,13 @@ impl AxisOp {
             Add(ix) => tensor.insert_axis(*ix),
             Rm(ix) => tensor.remove_axis(*ix),
             Move(from, to) => {
-                let mut permutation: Vec<usize> = (0..tensor.rank()).collect();
-                permutation.remove(*from);
-                permutation.insert(*to, *from);
-                let mut tmp = tensor.clone().permute_axes(&permutation)?;
+                let mut tmp = tensor.clone().move_axis(*from, *to)?;
                 std::mem::swap(tensor, &mut tmp);
                 Ok(())
             }
             Reshape(at, from, to) => {
                 let mut shape: TVec<usize> = tensor.shape().into();
-                self.change_shape_array(&mut shape, broadcasting)?;
+                self.change_shape_array(&mut shape, false)?;
                 if tensor.set_shape(&shape).is_ok() {
                     Ok(())
                 } else if broadcasting
@@ -464,7 +513,11 @@ impl TypedOp for AxisOp {
         Ok(tvec!(TypedFact::dt_shape(inputs[0].datum_type, shape)))
     }
 
-    fn invariants(&self, _inputs: &[&TypedFact], outputs: &[&TypedFact]) -> TractResult<Invariants> {
+    fn invariants(
+        &self,
+        _inputs: &[&TypedFact],
+        outputs: &[&TypedFact],
+    ) -> TractResult<Invariants> {
         let mut axes = vec![];
         for i in 0..outputs[0].rank() {
             if let Some(out) = self.transform_axis(i) {
@@ -487,7 +540,18 @@ impl TypedOp for AxisOp {
         if self.is_noop() {
             Ok(Some(TypedModelPatch::shunt_one_op(model, node)?))
         } else {
-            Ok(None)
+            let simplified = self.simplify();
+            if simplified.len() != 1 || &simplified[0] != self {
+                let mut patch = TypedModelPatch::default();
+                let mut wire = patch.tap_model(model, node.inputs[0])?;
+                for (ix, op) in simplified.into_iter().enumerate() {
+                    wire = patch.wire_node(format!("{}.{}", node.name, ix), op, &[wire])?[0];
+                }
+                patch.shunt_outside(model, node.id.into(), wire)?;
+                Ok(Some(patch))
+            } else {
+                Ok(None)
+            }
         }
     }
 
@@ -1173,4 +1237,44 @@ mod proptests {
         };
         pb.check().unwrap();
     }
+
+    #[test]
+    fn simplify_reshape() {
+        macro_rules! d {
+                ($($dim: expr),*) =>  { tvec!($($dim.to_dim()),*) }
+            }
+        assert_eq!(Reshape(3, d!(), d!()).simplify(), tvec!());
+        assert_eq!(Reshape(3, d!(2, 3), d!(2, 3)).simplify(), tvec!());
+        assert_eq!(Reshape(3, d!(1), d!()).simplify(), tvec!(Rm(3)));
+        assert_eq!(Reshape(3, d!(), d!(1)).simplify(), tvec!(Add(3)));
+        assert_eq!(
+            Reshape(3, d!(2, 3, 4), d!(2, 4, 3)).simplify(),
+            tvec!(Reshape(4, d!(3, 4), d!(4, 3)))
+        );
+        assert_eq!(
+            Reshape(3, d!(3, 4, 2), d!(4, 3, 2)).simplify(),
+            tvec!(Reshape(3, d!(3, 4), d!(4, 3)))
+        );
+        assert_eq!(
+            Reshape(3, d!(1, 2, 3), d!(3, 2)).simplify(),
+            tvec!(Rm(3), Reshape(3, d!(2, 3), d!(3, 2)))
+        );
+        assert_eq!(
+            Reshape(3, d!(2, 3), d!(1, 3, 2)).simplify(),
+            tvec!(Reshape(3, d!(2, 3), d!(3, 2)), Add(3))
+        );
+        assert_eq!(
+            Reshape(3, d!(2, 3, 1), d!(3, 2)).simplify(),
+            tvec!(Rm(5), Reshape(3, d!(2, 3), d!(3, 2)))
+        );
+        assert_eq!(
+            Reshape(3, d!(2, 3), d!(3, 2, 1)).simplify(),
+            tvec!(Add(5), Reshape(3, d!(2, 3), d!(3, 2)))
+        );
+        assert_eq!(
+            Reshape(2, d!(2, 2, 1), d!(4)).simplify(),
+            tvec!(Rm(4), Reshape(2, d!(2, 2), d!(4)))
+        );
+        assert_eq!(Reshape(1, d!(1, 2), d!(2)).simplify(), tvec!(Rm(1)));
+    }
 }
