@@ -1,6 +1,7 @@
 use super::proptest::*;
 use crate::ops::cnn::KernelFormat::*;
 use crate::ops::cnn::*;
+use crate::ops::math::round_ties_to_even;
 use crate::ops::matmul::*;
 use crate::ops::nn::DataFormat::*;
 use crate::ops::nn::*;
@@ -12,10 +13,6 @@ use tract_itertools::izip;
 use tract_ndarray::prelude::*;
 use tract_ndarray::*;
 
-fn round_away<F: num_traits::Float>(x: F) -> F {
-    x.abs().round() * x.signum()
-}
-
 pub fn qtensor(shape: Vec<usize>) -> BoxedStrategy<ArrayD<i8>> {
     let len = shape.iter().product::<usize>();
     vec(any::<i8>(), len..=len)
@@ -26,12 +23,12 @@ pub fn qtensor(shape: Vec<usize>) -> BoxedStrategy<ArrayD<i8>> {
 pub fn q_params() -> BoxedStrategy<MatMulQParams> {
     (-10i32..10, -10i32..10, -10i32..10, -3..3i32, -3..3i32, -3..3i32)
         .prop_map(|(a0, b0, c0, a_scale, b_scale, c_scale)| MatMulQParams {
-            a0: AttrOrInput::Attr(rctensor0(a0)),
-            b0: AttrOrInput::Attr(rctensor0(b0)),
-            c0: AttrOrInput::Attr(rctensor0(c0)),
-            a_scale: AttrOrInput::Attr(rctensor0(2f32.powi(a_scale))),
-            b_scale: AttrOrInput::Attr(rctensor0(2f32.powi(b_scale))),
-            c_scale: AttrOrInput::Attr(rctensor0(2f32.powi(c_scale))),
+            a0: tensor0(a0).into(),
+            b0: tensor0(b0).into(),
+            c0: tensor0(c0).into(),
+            a_scale: tensor0(2f32.powi(a_scale)).into(),
+            b_scale: tensor0(2f32.powi(b_scale)).into(),
+            c_scale: tensor0(2f32.powi(c_scale)).into(),
         })
         .boxed()
 }
@@ -113,7 +110,7 @@ impl QConvProblem {
             temp += &bias.clone().into_shape(shape).unwrap();
         }
         temp.mapv(|i| {
-            (round_away(i as f32 / scale as f32) as i32 + c0)
+            (round_ties_to_even(i as f32 / scale as f32) as i32 + c0)
                 .max(std::i8::MIN as i32)
                 .min(std::i8::MAX as i32) as i8
         })
@@ -332,7 +329,7 @@ fn a0_0() {
 #[test]
 fn scale_0() {
     let mut qp = MatMulQParams::noop_static(i8::datum_type());
-    qp.c_scale = AttrOrInput::Attr(rctensor0(9.274534f32));
+    qp.c_scale = tensor0(9.274534f32).into();
     QConvProblem {
         shape_in: HWC.from_n_c_hw(1, 1, &[1]).unwrap(),
         shape_out: HWC.from_n_c_hw(1, 1, &[1]).unwrap(),
@@ -351,7 +348,7 @@ fn scale_0() {
 #[test]
 fn scale_1() {
     let mut qp = MatMulQParams::noop_static(i8::datum_type());
-    qp.c_scale = AttrOrInput::Attr(rctensor0(1.1400417f32));
+    qp.c_scale = tensor0(1.1400417f32).into();
     QConvProblem {
         shape_in: HWC.from_n_c_hw(1, 1, &[1]).unwrap(),
         shape_out: HWC.from_n_c_hw(1, 1, &[1]).unwrap(),
@@ -370,7 +367,7 @@ fn scale_1() {
 #[test]
 fn scale_2() {
     let mut qp = MatMulQParams::noop_static(i8::datum_type());
-    qp.b_scale = AttrOrInput::Attr(rctensor0(0.5f32));
+    qp.b_scale = tensor0(0.5f32).into();
     QConvProblem {
         shape_in: HWC.from_n_c_hw(1, 1, &[1]).unwrap(),
         shape_out: HWC.from_n_c_hw(1, 1, &[1]).unwrap(),
@@ -386,6 +383,26 @@ fn scale_2() {
     .unwrap();
 }
 
+#[test]
+fn scale_3() {
+    let mut qp = MatMulQParams::noop_static(i8::datum_type());
+    qp.b_scale = tensor0(0.5f32).into();
+    qp.c_scale = tensor0(2f32).into();
+    QConvProblem {
+        shape_in: HWC.from_n_c_hw(1, 1, &[1]).unwrap(),
+        shape_out: HWC.from_n_c_hw(1, 1, &[1]).unwrap(),
+        kernel_format: OIHW,
+        group: 1,
+        data: arr2(&[[0i8]]).into_dyn(),
+        kernel: arr3(&[[[0i8]]]).into_dyn(),
+        bias: Some(arr2(&[[35i32]]).into_dyn()),
+        qp,
+        optim: true,
+    }
+    .check()
+    .unwrap();
+}
+
 #[test]
 fn group_0() {
     QConvProblem {
@@ -406,7 +423,7 @@ fn group_0() {
 #[test]
 fn group_1() {
     let mut qp = MatMulQParams::noop_static(i8::datum_type());
-    qp.b0 = AttrOrInput::Attr(rctensor0(1i32));
+    qp.b0 = tensor0(1i32).into();
     QConvProblem {
         shape_in: NCHW.from_n_c_hw(1, 2, &[1]).unwrap(),
         shape_out: NCHW.from_n_c_hw(1, 2, &[1]).unwrap(),
@@ -425,7 +442,7 @@ fn group_1() {
 #[test]
 fn group_2() {
     let mut qp = MatMulQParams::noop_static(i8::datum_type());
-    qp.b0 = AttrOrInput::Attr(rctensor0(1i32));
+    qp.b0 = tensor0(1i32).into();
     QConvProblem {
         shape_in: HWC.from_n_c_hw(1, 2, &[1]).unwrap(),
         shape_out: HWC.from_n_c_hw(1, 2, &[1]).unwrap(),
@@ -444,7 +461,7 @@ fn group_2() {
 #[test]
 fn rounding_on_arm() {
     let mut qp = MatMulQParams::noop_static(i8::datum_type());
-    qp.c_scale = AttrOrInput::Attr(rctensor0(1.3759452f32));
+    qp.c_scale = tensor0(1.3759452f32).into();
     QConvProblem {
         shape_in: HWC.from_n_c_hw(1, 1, &[1]).unwrap(),
         shape_out: HWC.from_n_c_hw(1, 2, &[1]).unwrap(),
