@@ -160,7 +160,10 @@ impl ConvUnary {
             let bias = bias.as_slice::<T>()?;
             ops.iter_mut().zip(bias.chunks(self.output_channels() / self.group)).for_each(
                 |(ops, bias)| {
-                    ops.push(ProtoFusedSpec::PerRowAdd(rctensor1(bias).into()));
+                    ops.push(ProtoFusedSpec::BinPerRow(
+                        rctensor1(bias).into(),
+                        tract_linalg::mmm::BinOp::Add,
+                    ));
                 },
             )
         }
@@ -185,12 +188,7 @@ impl ConvUnary {
         use crate::ops::matmul::mir_quant as qmm;
 
         let c_dt = self.q_params.as_ref().unwrap().0;
-        let b_dt = if let DatumType::QU8(qp) = b_dt {
-            let (zp, scale) = qp.zp_scale();
-            DatumType::QI8(QParams::ZpScale { zero_point: zp - 128, scale })
-        } else {
-            b_dt
-        };
+
         let params = self.q_params.as_ref().unwrap().1.as_outlet_ids(
             model,
             name,
@@ -507,7 +505,10 @@ impl ConvUnary {
     ) -> TractResult<OutletId> {
         let kernels = self.kernel_as_packed_as(&mmm.a_pack(k), m)?;
         let shape = kernels.shape();
-        let fused_ops = dispatch_copy!(Self::bias_as_non_linear(mmm.internal_type())(self))?;
+        let mut fused_ops = dispatch_copy!(Self::bias_as_non_linear(mmm.internal_type())(self))?;
+        for fo in &mut fused_ops {
+            fo.push(ProtoFusedSpec::Store);
+        }
         let mut iter = kernels.iter().cloned().zip(fused_ops.iter().cloned());
         let micro_ops = ArrayD::from_shape_fn(shape, |_| iter.next().unwrap());
 
@@ -828,7 +829,7 @@ impl TypedOp for ConvUnary {
         let one = 1.to_dim();
         Ok(tvec!(
             (
-                Cost::Params(inputs[0].datum_type),
+                Cost::Params(inputs[0].datum_type.unquantized()),
                 (self.kernel.len() + self.bias.as_ref().map(|b| b.len()).unwrap_or(0)).to_dim()
             ),
             (
