@@ -20,6 +20,7 @@ pub struct DeconvSum {
     pub input_shape: TVec<TDim>,
     pub adjustments: TVec<usize>,
     pub bias: Option<Arc<Tensor>>,
+    pub group: usize,
 }
 
 impl_dyn_hash!(DeconvSum);
@@ -53,7 +54,7 @@ impl EvalOp for DeconvSum {
             &self.pool_spec.dilations(),
             &self.pool_spec.strides(),
             &self.adjustments,
-        );
+        )?;
         let mut tensor = Tensor::zero::<f32>(&*output_shape.shape)?;
         let mut output = tensor.to_array_view_mut::<f32>()?;
         if let Some(b) = &self.bias {
@@ -99,7 +100,10 @@ impl EvalOp for DeconvSum {
                         {
                             let ocoord = ocoord.iter().map(|x| *x as usize).collect::<TVec<_>>();
                             let ocoord = self.pool_spec.data_format.from_n_c_hw(n, o, ocoord)?;
-                            output[&*ocoord.shape] += n_o_hkwk_hw[(n, o, kix, gix)];
+                            let value = n_o_hkwk_hw[(n, o, kix, gix)];
+                            if !value.is_nan() {
+                                output[&*ocoord.shape] += value
+                            }
                         }
                     }
                 }
