@@ -1,39 +1,41 @@
 use crate::internal::*;
 use ndarray::*;
 
-use tract_linalg::mmm::{FusedSpec, MatMatMul, MatrixStoreSpec, RoundingPolicy, ScratchSpace};
+use tract_linalg::mmm::{
+    BinOp, FusedSpec, InputStoreSpec, MatMatMul, OutputStore, OutputStoreSpec, RoundingPolicy,
+    ScratchSpace,
+};
 
 #[derive(PartialEq, Clone, Hash, Debug)]
 pub enum ProtoFusedSpec {
-    Min(AttrOrInput),
-    Max(AttrOrInput),
-    PerRowMul(AttrOrInput),
-    PerRowAdd(AttrOrInput),
-    PerColMul(AttrOrInput),
-    PerColAdd(AttrOrInput),
+    BinScalar(AttrOrInput, BinOp),
+    BinPerRow(AttrOrInput, BinOp),
+    BinPerCol(AttrOrInput, BinOp),
     AddRowColProducts(AttrOrInput, AttrOrInput),
-    ScalarMul(AttrOrInput),
-    ScalarAdd(AttrOrInput),
     AddUnicast(AttrOrInput),
     QScale(usize, RoundingPolicy, i32),
+    Store,
 }
 
 impl ProtoFusedSpec {
-    pub fn resolve<'t>(&'t self, inputs: &'t [Arc<Tensor>]) -> FusedSpec<'t> {
+    pub fn resolve<'t>(
+        &'t self,
+        inputs: &'t [Arc<Tensor>],
+        output_spec: &'t OutputStoreSpec,
+        output: OutputStore,
+    ) -> FusedSpec<'t> {
         match self {
-            ProtoFusedSpec::Min(v) => FusedSpec::Min(v.tensor(inputs)),
-            ProtoFusedSpec::Max(v) => FusedSpec::Max(v.tensor(inputs)),
-            ProtoFusedSpec::PerColAdd(v) => FusedSpec::PerColAdd(v.tensor(inputs)),
-            ProtoFusedSpec::PerRowAdd(v) => FusedSpec::PerRowAdd(v.tensor(inputs)),
-            ProtoFusedSpec::PerColMul(v) => FusedSpec::PerColMul(v.tensor(inputs)),
-            ProtoFusedSpec::PerRowMul(v) => FusedSpec::PerRowMul(v.tensor(inputs)),
-            ProtoFusedSpec::ScalarMul(v) => FusedSpec::ScalarMul(v.tensor(inputs)),
-            ProtoFusedSpec::ScalarAdd(v) => FusedSpec::ScalarAdd(v.tensor(inputs)),
+            ProtoFusedSpec::BinScalar(v, op) => FusedSpec::BinScalar(v.tensor(inputs), *op),
+            ProtoFusedSpec::BinPerRow(v, op) => FusedSpec::BinPerRow(v.tensor(inputs), *op),
+            ProtoFusedSpec::BinPerCol(v, op) => FusedSpec::BinPerCol(v.tensor(inputs), *op),
             ProtoFusedSpec::AddRowColProducts(row, col) => {
                 FusedSpec::AddRowColProducts(row.tensor(inputs), col.tensor(inputs))
             }
-            ProtoFusedSpec::AddUnicast(v) => FusedSpec::AddUnicast(v.tensor(inputs).view()),
+            ProtoFusedSpec::AddUnicast(v) => unsafe {
+                FusedSpec::AddUnicast(output_spec.wrap(&v.tensor(inputs).view()))
+            },
             ProtoFusedSpec::QScale(s, rp, m) => FusedSpec::QScale(*s, *rp, *m),
+            ProtoFusedSpec::Store => FusedSpec::Store(output),
         }
     }
 }
@@ -43,7 +45,7 @@ pub struct ConcreteMatMulGeometry {
     pub m: usize,
     pub k: usize,
     pub n: usize,
-    pub b_storage: MatrixStoreSpec,
+    pub b_storage: InputStoreSpec,
 }
 
 #[derive(Clone, Debug, Hash)]
@@ -69,13 +71,6 @@ impl ResolveTo<ConcreteMatMulGeometry> for SymbolicMatMulGeometry {
 pub type MatMulGeometry = GeometryBound<SymbolicMatMulGeometry, ConcreteMatMulGeometry>;
 
 impl MatMulGeometry {
-    fn m(&self) -> Cow<TDim> {
-        match self {
-            Self::Symbolic(it) => Cow::Borrowed(&it.m),
-            Self::Concrete(it) => Cow::Owned(it.m.to_dim()),
-        }
-    }
-
     fn k(&self) -> Cow<TDim> {
         match self {
             Self::Symbolic(it) => Cow::Borrowed(&it.k),
@@ -206,7 +201,8 @@ fn eval(
     c_final_shape: &[usize],
 ) -> TractResult<TVec<Arc<Tensor>>> {
     unsafe {
-        let a_dt = op.micro_ops.iter().next().unwrap().0.datum_type();
+        debug_assert!(op.micro_ops.len() > 0);
+        let size_of_a = (&*op.micro_ops.as_ptr()).0.datum_type().size_of();
         let mut c = Tensor::uninitialized_dt(op.c_fact.datum_type, &c_shape)?;
         let c_storage = op.mmm.c_view_with_axis(c_m_axis, c_n_axis);
         if op
@@ -231,33 +227,30 @@ fn eval(
                     c_view.offset_axis_unchecked(ix, dim as isize);
                 }
                 let (pa, fused) = ops.iter().next().unwrap();
-                let f: Vec<FusedSpec> = fused.iter().map(|f| f.resolve(inputs)).collect::<Vec<_>>();
-                op.mmm.run_with_scratch_space(
-                    geometry.m,
-                    geometry.k,
-                    geometry.n,
-                    scratch,
-                    &op.mmm.a_packed(a_dt.size_of(), geometry.k).wrap(&pa.view()),
-                    &geometry
+                let c_store = c_storage.wrap(&c_view);
+                let mut f = tvec!(FusedSpec::AddMatMul {
+                    k: geometry.k,
+                    a: op.mmm.a_packed(size_of_a, geometry.k).wrap(&pa.view()),
+                    b: geometry
                         .b_storage
                         .wrap(&TensorView::at_prefix_unchecked(&inputs[0], &*b_prefix)),
-                    &mut c_storage.wrap(&c_view),
-                    &f,
-                )?;
+                });
+                f.extend(fused.iter().map(|f| f.resolve(inputs, &c_storage, c_store)));
+                op.mmm.run_with_scratch_space(geometry.m, geometry.n, scratch, &f)?;
             }
         } else {
-            let (pa, fused) = op.micro_ops.iter().next().unwrap();
-            let f: Vec<FusedSpec> = fused.iter().map(|f| f.resolve(inputs)).collect::<Vec<_>>();
-            op.mmm.run_with_scratch_space(
-                geometry.m,
-                geometry.k,
-                geometry.n,
-                scratch,
-                &op.mmm.a_packed(a_dt.size_of(), geometry.k).wrap(&pa.view()),
-                &geometry.b_storage.wrap(&inputs[0].view()),
-                &mut c_storage.wrap(&c.view_mut()),
-                &f,
-            )?;
+            let (pa, fused) = &*op.micro_ops.as_ptr();
+            let c_store = c_storage.wrap(&c.view_mut());
+            let mut f = Vec::with_capacity(fused.len() + 1);
+            f.push(FusedSpec::AddMatMul {
+                k: geometry.k,
+                a: op.mmm.a_packed(size_of_a, geometry.k).wrap(&pa.view()),
+                b: geometry.b_storage.wrap(&inputs[0].view()),
+            });
+            for ix in 0..fused.len() {
+                f.push(fused.get_unchecked(ix).resolve(inputs, &c_storage, c_store));
+            }
+            op.mmm.run_with_scratch_space(geometry.m, geometry.n, scratch, &f)?;
         }
         c.set_shape_unchecked(c_final_shape);
         Ok(tvec!(c.into_arc_tensor()))
@@ -284,7 +277,7 @@ impl TypedOp for LirMatMulUnary {
         Ok(tvec!(
             (Cost::FMA(self.mmm.internal_type()), sums * self.geometry.k().as_ref()),
             (
-                Cost::Params(self.micro_ops.as_slice().unwrap()[0].0.datum_type()),
+                Cost::Params(self.micro_ops.as_slice().unwrap()[0].0.datum_type().unquantized()),
                 self.micro_ops.iter().fold(0.to_dim(), |sum, a| sum + a.0.len())
             )
         ))
@@ -317,34 +310,12 @@ impl TypedOp for LirMatMulUnary {
             }
         }
 
-        let merge = |fused_micro_op: &ArrayD<Vec<ProtoFusedSpec>>,
-                     additional_inputs: &[OutletId]|
-         -> TractResult<Option<TypedModelPatch>> {
-            let mut new_op = self.clone();
-            new_op
-                .micro_ops
-                .zip_mut_with(fused_micro_op, |lhs, rhs| lhs.1.extend(rhs.iter().cloned()));
-            let mut patch = TypedModelPatch::new(format!("fusing {}", succ));
-            patch.dont_apply_twice = Some(format!("Fuse {} into {}", succ.name, node.name));
-            let inputs = node
-                .inputs
-                .iter()
-                .chain(additional_inputs.iter())
-                .map(|i| patch.tap_model(model, *i))
-                .collect::<TractResult<TVec<OutletId>>>()?;
-            let output = patch.wire_node(&node.name, new_op, &inputs)?;
-            patch.shunt_outside(model, succ.id.into(), output[0])?;
-            Ok(Some(patch))
-        };
-
-        let merge_broadcast = |spec: &[ProtoFusedSpec], additional_inputs: &[OutletId]| {
-            let array = arr0(spec.to_vec()).into_dyn();
-            merge(&array, additional_inputs)
-        };
-
         if let Some(op) = succ.op_as::<ops::element_wise::ElementWiseOp>().map(|ew| ew.0.as_ref()) {
             if let Some(cast) = op.downcast_ref::<ops::cast::Cast>().map(|cast| cast.to) {
-                if cast == i8::datum_type() && self.c_fact.datum_type == i32::datum_type() {
+                if (cast.unquantized() == i8::datum_type()
+                    || cast.unquantized() == u8::datum_type())
+                    && self.c_fact.datum_type == i32::datum_type()
+                {
                     let at = self.micro_ops.iter().nth(0).unwrap().0.datum_type();
                     let bt = model.outlet_fact(node.inputs[0])?.datum_type;
                     let mmm = tract_linalg::ops()
@@ -358,7 +329,7 @@ impl TypedOp for LirMatMulUnary {
                         )
                         .unwrap();
 
-                    let c_fact = TypedFact::dt_shape(i8::datum_type(), self.c_fact.shape.clone());
+                    let c_fact = TypedFact::dt_shape(cast, self.c_fact.shape.clone());
                     let mut patch = TypedModelPatch::fuse_with_next(
                         model,
                         &node,
@@ -368,72 +339,29 @@ impl TypedOp for LirMatMulUnary {
                     return Ok(Some(patch));
                 }
             } else if let Some(op) = op.downcast_ref::<ops::math::QScale>() {
-                return merge_broadcast(
+                return self.fuse_op_with_broadcast(
+                    model,
+                    node,
                     &[ProtoFusedSpec::QScale(op.shift, op.policy, op.mult)],
                     &[],
                 );
             }
         } else if let Some(op) = succ.op_as::<ops::binary::UnaryOp>() {
-            if op.a.len() == 1 {
-                if op.mini_op.is::<ops::math::Max>() {
-                    return merge_broadcast(&[ProtoFusedSpec::Max((&op.a).into())], &[]);
-                } else if op.mini_op.is::<ops::math::Min>() {
-                    return merge_broadcast(&[ProtoFusedSpec::Min((&op.a).into())], &[]);
-                } else if op.mini_op.is::<ops::math::Mul>() {
-                    return merge_broadcast(&[ProtoFusedSpec::ScalarMul((&op.a).into())], &[]);
-                } else {
-                    return Ok(None);
-                }
-            }
-            let mut arg = op.a.clone().into_tensor();
-            for axis_change in self.reshape_post.iter().rev() {
-                axis_change.recip().change_tensor(&mut arg, true)?;
-            }
-            if arg.shape()[self.c_n_axis] == 1
-                && arg.shape()[self.c_m_axis].to_dim() == self.c_fact.shape[self.c_m_axis]
-                && (op.mini_op.is::<ops::math::Mul>() || op.mini_op.is::<ops::math::Add>())
-            {
-                let prefix = arg
-                    .shape()
-                    .iter()
-                    .enumerate()
-                    .filter_map(|(ix, d)| {
-                        if ix != self.c_n_axis && ix != self.c_m_axis {
-                            Some(*d)
-                        } else {
-                            None
-                        }
-                    })
-                    .collect::<TVec<_>>();
-
-                assert_eq!(
-                    prefix.iter().cloned().product::<usize>()
-                        * self.geometry.m().to_usize().unwrap(),
-                    arg.len()
-                );
-
-                let arg_len = arg.len();
-                let arg = arg.into_shape(&[arg_len])?;
-                let mut i = 0;
-                let arg = ArrayD::from_shape_simple_fn(&*prefix, || {
-                    let t = arg
-                        .slice(
-                            0,
-                            i * self.geometry.m().to_usize().unwrap(),
-                            (i + 1) * self.geometry.m().to_usize().unwrap(),
-                        )
-                        .unwrap();
-                    i += 1;
-                    if op.mini_op.is::<ops::math::Mul>() {
-                        vec![ProtoFusedSpec::PerRowMul(t.into())]
-                    } else if op.mini_op.is::<ops::math::Add>() {
-                        vec![ProtoFusedSpec::PerRowAdd(t.into())]
-                    } else {
-                        unreachable!()
-                    }
-                });
-                return merge(&arg, &[]);
+            let binop =
+                if let Some(op) = op.mini_op.as_linalg_binop() { op } else { return Ok(None) };
+            let shape = op.a.shape().into();
+            return self.fuse_binary(model, node, &shape, op.a.clone().into(), binop, &[]);
+        } else if let Some(op) = succ.op_as::<ops::binary::TypedBinOp>() {
+            let mut binop =
+                if let Some(op) = op.0.as_linalg_binop() { op } else { return Ok(None) };
+            let flipped = succ.inputs[0].node == node.id;
+            if flipped {
+                binop = binop.flip();
             }
+            let other_outlet = succ.inputs[flipped as usize];
+            let other_fact = model.outlet_fact(other_outlet)?;
+            let value = node.inputs.len().into();
+            return self.fuse_binary(model, node, &other_fact.shape, value, binop, &[other_outlet]);
         } else if let Some(op) = succ.op_as::<ops::binary::MergeOpUnicast>() {
             if self.c_n_axis == self.c_final_shape.rank() - 2
                 && self.c_m_axis == self.c_final_shape.rank() - 1
@@ -443,7 +371,9 @@ impl TypedOp for LirMatMulUnary {
                 let other_input = succ.inputs[other_slot];
 
                 if op.0.is::<ops::math::Add>() {
-                    return merge_broadcast(
+                    return self.fuse_op_with_broadcast(
+                        model,
+                        node,
                         &[ProtoFusedSpec::AddUnicast(node.inputs.len().into())],
                         &[other_input],
                     );
@@ -455,3 +385,89 @@ impl TypedOp for LirMatMulUnary {
 
     as_op!();
 }
+
+impl LirMatMulUnary {
+    fn fuse_op(
+        &self,
+        model: &TypedModel,
+        node: &TypedNode,
+        fused_micro_op: &ArrayD<Vec<ProtoFusedSpec>>,
+        additional_inputs: &[OutletId],
+    ) -> TractResult<Option<TypedModelPatch>> {
+        let succ = model.node(node.outputs[0].successors[0].node);
+        let mut new_op = self.clone();
+        new_op.micro_ops.zip_mut_with(fused_micro_op, |lhs, rhs| {
+            lhs.1.pop();
+            lhs.1.extend(rhs.iter().cloned());
+            lhs.1.push(ProtoFusedSpec::Store);
+        });
+        let mut patch = TypedModelPatch::new(format!("fusing {}", succ));
+        patch.dont_apply_twice = Some(format!("Fuse {} into {}", succ.name, node.name));
+        let inputs = node
+            .inputs
+            .iter()
+            .chain(additional_inputs.iter())
+            .map(|i| patch.tap_model(model, *i))
+            .collect::<TractResult<TVec<OutletId>>>()?;
+        let output = patch.wire_node(&node.name, new_op, &inputs)?;
+        patch.shunt_outside(model, succ.id.into(), output[0])?;
+        Ok(Some(patch))
+    }
+
+    fn fuse_op_with_broadcast(
+        &self,
+        model: &TypedModel,
+        node: &TypedNode,
+        fused_micro_op: &[ProtoFusedSpec],
+        additional_inputs: &[OutletId],
+    ) -> TractResult<Option<TypedModelPatch>> {
+        let array = arr0(fused_micro_op.to_vec()).into_dyn();
+        self.fuse_op(model, node, &array, additional_inputs)
+    }
+
+    fn fuse_binary(
+        &self,
+        model: &TypedModel,
+        node: &TypedNode,
+        shape: &ShapeFact,
+        value: AttrOrInput,
+        binop: BinOp,
+        additional_inputs: &[OutletId],
+    ) -> TractResult<Option<TypedModelPatch>> {
+        if shape.volume() == 1.to_dim() {
+            return self.fuse_op_with_broadcast(
+                model,
+                node,
+                &[ProtoFusedSpec::BinScalar(value, binop)],
+                additional_inputs,
+            );
+        }
+        let mut other_shape = shape.to_owned();
+        for axis_change in self.reshape_post.iter().rev() {
+            if axis_change.recip().change_shape(&mut other_shape, true).is_err() {
+                return Ok(None)
+            }
+        }
+        if other_shape[self.c_m_axis] == self.c_fact.shape[self.c_m_axis]
+            && other_shape[self.c_m_axis] == other_shape.volume()
+        {
+            return self.fuse_op_with_broadcast(
+                model,
+                node,
+                &[ProtoFusedSpec::BinPerRow(value, binop)],
+                additional_inputs,
+            );
+        }
+        if other_shape[self.c_n_axis] == self.c_fact.shape[self.c_n_axis]
+            && other_shape[self.c_n_axis] == other_shape.volume()
+        {
+            return self.fuse_op_with_broadcast(
+                model,
+                node,
+                &[ProtoFusedSpec::BinPerCol(value, binop)],
+                additional_inputs,
+            );
+        }
+        Ok(None)
+    }
+}
