@@ -185,6 +185,7 @@ pub trait TypedOp:
         model: &TypedModel,
         node: &TypedNode,
         patch: &mut TypedModelPatch,
+        slice_suffix: &str,
         output_slot: usize,
         axis: usize,
         start: usize,
@@ -197,7 +198,7 @@ pub trait TypedOp:
         } else {
             let wire = patch.tap_model(model, outlet)?;
             let wire = patch.wire_node(
-                &format!("{}-{}-slice-{}-{}..{}", node.name, output_slot, axis, start, end),
+                &format!("{}-{}.{}", node.name, output_slot, slice_suffix),
                 crate::ops::array::Slice { start: start.to_dim(), axis, end: end.to_dim() },
                 &[wire],
             )?[0];
